#define __KERNEL__
#define MODULE

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/netdevice.h>
#include <net/tcp.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

#include <linux/kernel.h>
#include <linux/skbuff.h>

#define SNUFFI_MAJOR_VERSION 	0
#define SNUFFI_MINOR_VERSION	1

MODULE_LICENSE("GPL");

/* Globals */
struct nf_hook_ops snuffi_in_hook_struct;
struct nf_hook_ops snuffi_out_hook_struct;

/* Prototypes */
unsigned int snuffi_in_hook(unsigned int hooknum, struct sk_buff **skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *));
unsigned int snuffi_out_hook(unsigned int hooknum, struct sk_buff **skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *));
int print_ipv4(struct iphdr *iph);
int print_tcp(struct tcphdr *tcph);

/* Function snuffi_in_hook */
unsigned int snuffi_in_hook(unsigned int hooknum, struct sk_buff **skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *)) {

	printk("[SNUFFI] - In \"%s\" -\n",
			in->name);
			
	/* IPv4 */

	print_ipv4((struct iphdr *) (*skb)->nh.iph);

	/* Check protocol field for transport layer protocol*/
	
	if((*skb)->nh.iph->protocol == IPPROTO_TCP && (*skb)->nh.iph->ihl == 5) 
		print_tcp((struct tcphdr *) (*skb)->h.th);	
	
	return NF_ACCEPT;	
}

/* Function snuffi_out_hook */
unsigned int snuffi_out_hook(unsigned int hooknum, struct sk_buff **skb, const struct net_device *in, const struct net_device *out, int (*okfn)(struct sk_buff *)) {

        printk("[SNUFFI] - Out \"%s\" -\n",
			out->name);

        /* IPv4 */

        print_ipv4((struct iphdr *) (*skb)->nh.iph);

        /* Check protocol field for transport layer protocol*/

        if((*skb)->nh.iph->protocol == IPPROTO_TCP && (*skb)->nh.iph->ihl == 5)
                print_tcp((struct tcphdr *) (*skb)->h.th);

        return NF_ACCEPT;
}

/* Function print_ipv4 */
int print_ipv4(struct iphdr *iph) {

	if(iph == NULL)
		return(1);
	
	printk("[SNUFFI] IPv4 - Header length: %u, TOS: 0x%x\n",
			iph->ihl, iph->tos);
        printk("[SNUFFI] IPv4 - Total Length: %u, ID: %u\n",
			ntohs(iph->tot_len), ntohs(iph->id));
        printk("[SNUFFI] IPv4 - Frag. Offset: 0x%04x, TTL: %u\n",
			ntohs(iph->frag_off), iph->ttl);
        printk("[SNUFFI] IPv4 - Protocol: %u, Sum: 0x%04x\n",
			iph->protocol, ntohs(iph->check));
        printk("[SNUFFI] IPv4 - Src: %u.%u.%u.%u\n",
			NIPQUAD(iph->saddr));
        printk("[SNUFFI] IPv4 - Dst: %u.%u.%u.%u\n",
			NIPQUAD(iph->daddr));

	return(0);
};

/* Function print_tcp */
int print_tcp(struct tcphdr *tcph) {

	if(tcph == NULL)
		return(1);
	
	printk("[SNUFFI] TCP - Src Port: %u, Dst Port: %u\n",
			ntohs(tcph->source), ntohs(tcph->dest));
	printk("[SNUFFI] TCP - Seq.: %u\n",
			ntohl(tcph->seq));
	printk("[SNUFFI] TCP - Ack-Seq.: %u\n",
			ntohl(tcph->ack_seq));
	printk("[SNUFFI] TCP - Data Off.: %u, Flags: ",
			tcph->doff);
	if(tcph->fin) 
		printk("F");
        if(tcph->syn)
	        printk("S");
        if(tcph->rst)
                printk("R");
        if(tcph->psh)
                printk("P");
        if(tcph->ack)
                printk("A");
        if(tcph->urg)
                printk("U");
        if(tcph->ece)
                printk("E");
        if(tcph->cwr)
                printk("C");
	printk("\n[SNUFFI] TCP - Window: %u, Sum: 0x%04x\n",
			ntohs(tcph->window), ntohs(tcph->check));
	printk("[SNUFFI] TCP - Urg-Ptr.: 0x%04x\n",
			ntohs(tcph->urg_ptr));

	return(0);
}

/* Function init_module */
int init_module(void) {
	int ret;

	snuffi_in_hook_struct.list.next	= NULL;
	snuffi_in_hook_struct.list.prev = NULL;
	snuffi_in_hook_struct.pf	= PF_INET;	/* IPv4 Family */
	snuffi_in_hook_struct.hook 	= snuffi_in_hook;/* Called hook */
	snuffi_in_hook_struct.hooknum 	= NF_IP_LOCAL_IN;/* Incoming traffic */
	
        snuffi_out_hook_struct.list.next = NULL;
	snuffi_out_hook_struct.list.prev = NULL;
	snuffi_out_hook_struct.pf        = PF_INET;      /* IPv4 Family */
	snuffi_out_hook_struct.hook      = snuffi_out_hook;/* Called hook */
	snuffi_out_hook_struct.hooknum   = NF_IP_LOCAL_OUT;/*Outgoing traffic*/
					
	printk("[SNUFFI] Kernel module(v%u.%u) for IPv4 sniffing with netfilter by Maik Pfeil(arbon@elxsi.de).\n", 
			SNUFFI_MAJOR_VERSION, SNUFFI_MINOR_VERSION);

	printk("[SNUFFI] DEBUG - nf_register_hook.\n");
	if(ret = nf_register_hook(&snuffi_in_hook_struct))
		printk("[SNUFFI] DEBUG - Can't nf_register_hook module for incoming queue.\n");
	
        if(ret = nf_register_hook(&snuffi_out_hook_struct))
	        printk("[SNUFFI] DEBUG - Can't nf_register_hook module for outgoing queue.\n");
	
	return ret;
}

/* Function cleanup_module */
void cleanup_module(void) {
	printk("[SNUFFI] DEBUG - nf_unregister_hook.\n");
	
        nf_unregister_hook(&snuffi_out_hook_struct);
	nf_unregister_hook(&snuffi_in_hook_struct);
}

/*
 * gcc -O2 -Wall -I/usr/src/linux/include -c snuffi.c
 * insmod snuffi.o
 */
